from pathlib import Path  # noqa: INP001
from typing import Any, Dict

import numpy as np
from numpy.typing import NDArray
from ropt.enums import ConstraintType, EventType
from ropt.evaluator import EvaluatorContext, EvaluatorResult
from ropt.optimization import Event
from ropt.plan import OptimizationPlanRunner
from ropt.results import FunctionResults
from ruamel import yaml

# For convenience we use a YAML file to store the optimizer options:
options = yaml.YAML(typ="safe", pure=True).load(Path("discrete_ga.yml"))

CONFIG: Dict[str, Any] = {
    "variables": {
        # Ignored, but needed to establish the number of variables:
        "initial_values": 2 * [0.0],
        "lower_bounds": [0.0, 0.0],
        "upper_bounds": [10.0, 10.0],
    },
    "optimizer": {
        "method": "soo.nonconvex.ga.GA",
        "options": options,
    },
    "nonlinear_constraints": {
        "types": [ConstraintType.LE],
        "rhs_values": [0.0],
    },
}


def function(variables: NDArray[np.float64], _: EvaluatorContext) -> EvaluatorResult:
    x, y = variables[0, :]
    objectives = np.array(-min(3 * x, y), ndmin=2, dtype=np.float64)
    constraints = np.array(x + y - 10, ndmin=2, dtype=np.float64)
    return EvaluatorResult(objectives=objectives, constraints=constraints)


def report(event: Event) -> None:
    assert event.results is not None
    for item in event.results:
        if isinstance(item, FunctionResults):
            assert item.functions is not None
            print(f"evaluation: {item.result_id}")
            print(f"  variables: {item.evaluations.variables}")
            print(f"  objective: {item.functions.weighted_objective}\n")


def run_optimization(config: Dict[str, Any]) -> None:
    optimal_result = (
        OptimizationPlanRunner(config, function)
        .add_observer(EventType.FINISHED_EVALUATION, report)
        .run()
        .results
    )
    assert optimal_result is not None
    assert optimal_result.functions is not None
    print(f"BEST RESULT: {optimal_result.result_id}")
    print(f"  variables: {optimal_result.evaluations.variables}")
    print(f"  objective: {optimal_result.functions.weighted_objective}\n")


def main() -> None:
    run_optimization(CONFIG)


if __name__ == "__main__":
    main()



from __future__ import annotations

import importlib
import inspect
from typing import Any, Callable, Dict, Optional, Tuple, Union

from pydantic import BaseModel, ConfigDict, Field
from pymoo.core.algorithm import Algorithm  # noqa: TCH002
from pymoo.core.operator import Operator  # noqa: TCH002
from pymoo.core.termination import Termination  # noqa: TCH002
from pymoo.termination import get_termination


class _ParametersBaseModel(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        str_min_length=1,
        str_strip_whitespace=True,
        validate_default=True,
        frozen=True,
    )


ParameterValues = Union[bool, int, float, str]


class ObjectConfig(_ParametersBaseModel):
    """Configuration for an operator.

    Configures a pymoo object, which can be used a a parameter. They are
    specified by their fully qualified name within the `pymoo` module. For
    example to specify the `FloatRandomSampling` operator, the operator should
    be specified as 'operators.sampling.rnd.FloatRandomSampling`.

    The parameters are specified as a dict that will be passed as keyword
    arguments to the operator. These may consist of the basic types, such as
    booleans, integers, floats and strings, or as a nested object specification.

    Attributes:
        object:     The fully qualified operator class name.
        parameters: The parameters passed to the operator constructor.
    """

    object_: str = Field(alias="object")
    parameters: Dict[str, Union[ParameterValues, ObjectConfig]] = {}


class TerminationConfig(_ParametersBaseModel):
    """Configuration of termination conditions.

    Termination objects are used within `pymoo` to determine when an
    optimization algorithm should terminate. They are specified by a string for
    a set of defaults, which are specified here by there fully qualified name
    within the `pymoo.termination` module. For example to specify the
    `MaximumGenerationTermination` termination, the it should be specified here
    as 'max_gen.MaximumGenerationTermination`.

    In addition to the fully qualified name it is also possible to use the short
    name of the termination object, as it is used by the `get_termination`
    function, e.g. `n_eval`.

    For details about termination objects, consult the `pymoo` manual:
    [`Operators`](https://pymoo.org/interface/termination.html)


    Note:
        Instead of using a termination object, it is also possible to use a
        tuple of termination conditions. See the `pymoo` manual for details.

    Attributes:
        name:       The fully qualified termination class name.
        parameters: The parameters passed to the termination constructor.
    """

    name: str = "soo"
    parameters: Dict[str, Union[int, float]] = {}


class ConstraintsConfig(_ParametersBaseModel):
    """Configuration of constraint handling.

    Constraint objects are used within `pymoo` to determine how constraints
    should be handled. Here they are specified by a fully qualified name within
    the `pymoo.constraints` module. For example to specify the
    `ConstraintsAsPenalty` object, the object should be specified
    here as 'as_penalty.ConstraintsAsPenalty`.

    For details about constraint handling, consult the `pymoo` manual:
    [`Operators`](https://pymoo.org/constraints/index.html)

    Attributes:
        name:       The fully qualified constraint class name.
        parameters: The parameters passed to the constraint constructor.
    """

    name: str
    parameters: Dict[str, Union[bool, int, float, str]] = {}


class ParametersConfig(_ParametersBaseModel):
    """Configuration of the `pymoo` algorithm.

    Optimization by `pymoo` algorithms requires three objects:

    1. An algorithm object. Here they are specified by a fully qualified name
       within the `pymoo.algorithms` module. For example to specify the `GA`
       algorithm, it  should be specified here as 'soo.nonconvex.ga.GA`. The
       algorithm is initialized by parameters that here are specified by the
       `parameters` field.
    2. A constraints handling object.
    3. A termination object, or a tuple of termination values.
    4. A seed value for the random number generator used by the algorithms.

    Attributes:
        algorithm:   The fully qualified name of the algorithm class
        parameters:  The parameters passed to the algorithm constructor
        constraints: Specification of the constraint handling object to use
        termination: Specification of the termination object to use
        seed:        The seed value for the random number generator
    """

    algorithm: Union[str, Algorithm]
    parameters: Dict[str, Union[ParameterValues, ObjectConfig]] = {}
    constraints: Optional[ConstraintsConfig] = None
    termination: Union[TerminationConfig, Tuple[Any, ...]] = TerminationConfig()
    seed: int = 1

    def get_algorithm(self) -> Algorithm:
        """Parse the algorithm config and its parameters.

        Raises:
            ValueError: When the algorithm object is not found.

        Returns:
            A `pymoo` algorithm object.
        """
        parameters = _parse_parameters(self.parameters)
        algorithm = _get_class(
            self.algorithm,
            prefix="pymoo.algorithms",
            keyword="algorithm",
        )
        if algorithm is None:
            msg = f"Algorithm not found: {algorithm}"
            raise ValueError(msg)
        return algorithm(**parameters)

    def get_termination(self) -> Union[Termination, Tuple[Any, ...]]:
        """Parse the termination config.

        Raises:
            ValueError: When the termination object is not found.

        Returns:
            A `pymoo` termination object.
        """
        if isinstance(self.termination, tuple):
            return self.termination
        module_name, _, name = self.termination.name.rpartition(".")
        if not module_name:
            try:
                return get_termination(name, **self.termination.parameters)
            except Exception as exc:
                msg = f"Failed to run get_termination for: {name}"
                raise ValueError(msg) from exc
        termination_class = _get_class(
            self.termination.name,
            prefix="pymoo.termination",
            keyword="name",
        )
        if termination_class is None:
            msg = f"Termination class not found: {termination_class}"
            raise ValueError(msg)
        return termination_class(**self.termination.parameters)

    def get_constraints(self) -> Any:  # noqa: ANN401
        """Parse the constraints config.

        Raises:
            ValueError: When the constraints object is not found.

        Returns:
            A `pymoo` constraints class.
        """
        if self.constraints is None:
            return None
        constraints_class = _get_class(
            self.constraints.name,
            prefix="pymoo.constraints",
            keyword="name",
        )
        if constraints_class is None:
            msg = f"Constraints class not found: {constraints_class}"
            raise ValueError(msg)
        return constraints_class


def _get_class(name: str, prefix: str, keyword: str) -> Optional[Callable[..., Any]]:
    module_name, _, class_name = name.rpartition(".")
    if not module_name:
        msg = f"Not a properly formatted {keyword} name: {name}"
        raise ValueError(msg)
    full_module_name = prefix + "." + module_name
    try:
        module = importlib.import_module(full_module_name)
    except ImportError as exc:
        msg = f"Invalid value for {keyword}: {full_module_name}"
        raise ValueError(msg) from exc
    return next(
        (
            cls
            for _, cls in (inspect.getmembers(module, inspect.isclass))
            if cls.__name__ == class_name
        ),
        None,
    )


def _parse_parameters(
    parameters: Dict[str, Union[ParameterValues, ObjectConfig]],
) -> Dict[str, Union[ParameterValues, Operator]]:
    parsed = {}
    for key, value in parameters.items():
        if isinstance(value, ObjectConfig):
            object_class = _get_class(value.object_, prefix="pymoo", keyword="object")
            if object_class is None:
                msg = f"Object class not found: {object}"
                raise ValueError(msg)
            operator_parameters = _parse_parameters(value.parameters)
            parsed[key] = object_class(**operator_parameters)
        else:
            parsed[key] = value
    return parsed



from typing import Any, Dict  # noqa: INP001

import numpy as np
from numpy.typing import NDArray
from ropt.enums import EventType
from ropt.evaluator import EvaluatorContext, EvaluatorResult
from ropt.optimization import Event
from ropt.plan import OptimizationPlanRunner
from ropt.results import FunctionResults

CONFIG: Dict[str, Any] = {
    "variables": {
        "initial_values": 2 * [0.0],
        "lower_bounds": [0.0, 0.0],
        "upper_bounds": [2.0, 2.0],
    },
    "optimizer": {
        "method": "soo.nonconvex.nelder.NelderMead",
        "options": {
            "termination": ("n_iter", 30),
        },
    },
}


def rosenbrock(variables: NDArray[np.float64], _: EvaluatorContext) -> EvaluatorResult:
    objectives = np.zeros((variables.shape[0], 1), dtype=np.float64)
    for idx in range(variables.shape[0]):
        x, y = variables[idx, :]
        objectives[idx, 0] = (1.0 - x) ** 2 + 100 * (y - x * x) ** 2
    return EvaluatorResult(objectives=objectives)


def report(event: Event) -> None:
    assert event.results is not None
    for item in event.results:
        if isinstance(item, FunctionResults):
            assert item.functions is not None
            print(f"evaluation: {item.result_id}")
            print(f"  variables: {item.evaluations.variables}")
            print(f"  objective: {item.functions.weighted_objective}\n")


def run_optimization(config: Dict[str, Any]) -> None:
    optimal_result = (
        OptimizationPlanRunner(config, rosenbrock)
        .add_observer(EventType.FINISHED_EVALUATION, report)
        .run()
        .results
    )
    assert optimal_result is not None
    assert optimal_result.functions is not None
    print(f"BEST RESULT: {optimal_result.result_id}")
    print(f"  variables: {optimal_result.evaluations.variables}")
    print(f"  objective: {optimal_result.functions.weighted_objective}\n")


def main() -> None:
    run_optimization(CONFIG)


if __name__ == "__main__":
    main()



import importlib
import os
import shutil
from pathlib import Path
from typing import Any, Optional

EXAMPLES_DIR = Path(__file__).parent.parent / "examples"


def _load_from_file(name: str, sub_path: Optional[str] = None) -> Any:
    path = EXAMPLES_DIR
    if sub_path is not None:
        path = path / sub_path
    path = path / f"{name}.py"
    spec = importlib.util.spec_from_file_location(name, path)
    assert spec is not None
    module = importlib.util.module_from_spec(spec)
    assert spec.loader is not None
    spec.loader.exec_module(module)
    return module


def test_rosenbrock_de(tmp_path: Path) -> None:
    os.chdir(tmp_path)
    module = _load_from_file("rosenbrock_nm")
    module.main()


def test_discrete_ga(tmp_path: Path) -> None:
    shutil.copyfile(
        EXAMPLES_DIR / "discrete_ga.yml",
        tmp_path / "discrete_ga.yml",
    )
    os.chdir(tmp_path)
    module = _load_from_file("discrete_ga")
    module.main()



"""This module implements the pymoo optimization plugin."""

from __future__ import annotations

import copy
import importlib
import inspect
import os
import sys
from contextlib import redirect_stdout
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Final,
    Optional,
    TextIO,
    Tuple,
    Union,
)

import numpy as np
from pymoo.core.problem import Problem
from pymoo.optimize import minimize
from ropt.enums import ConstraintType
from ropt.plugins.optimizer.base import Optimizer, OptimizerCallback, OptimizerPlugin
from ropt.plugins.optimizer.utils import create_output_path, filter_linear_constraints

from ._config import ParametersConfig

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from ropt.config.enopt import EnOptConfig

_OUTPUT_FILE: Final = "optimizer_output"

# These algorithms do not allow NaN fucntion values:
_NO_FAILURE_HANDLING: Final = {"NelderMead"}


@dataclass
class _Constraints:
    linear_eq: NDArray[np.bool_] = field(
        default_factory=lambda: np.array([], dtype=np.bool_)
    )
    linear_ineq: NDArray[np.bool_] = field(
        default_factory=lambda: np.array([], dtype=np.bool_)
    )
    nonlinear_eq: NDArray[np.bool_] = field(
        default_factory=lambda: np.array([], dtype=np.bool_)
    )
    nonlinear_ineq: NDArray[np.bool_] = field(
        default_factory=lambda: np.array([], dtype=np.bool_)
    )
    coefficients: NDArray[np.float64] = field(
        default_factory=lambda: np.array([]),
    )
    rhs_values: NDArray[np.float64] = field(default_factory=lambda: np.array([]))
    ineq_func: Optional[Callable[[NDArray[np.float64]], NDArray[np.float64]]] = None
    eq_func: Optional[Callable[[NDArray[np.float64]], NDArray[np.float64]]] = None


class _Problem(Problem):  # type: ignore
    def __init__(  # noqa: PLR0913
        self,
        n_var: int,
        lower: NDArray[np.float64],
        upper: NDArray[np.float64],
        function: Callable[[NDArray[np.float64]], NDArray[np.float64]],
        constraints: _Constraints,
        *,
        parallel: bool = True,
    ) -> None:
        n_ieq_constr = np.count_nonzero(constraints.linear_ineq)
        n_ieq_constr += np.count_nonzero(constraints.nonlinear_ineq)
        n_eq_constr = np.count_nonzero(constraints.linear_eq)
        n_eq_constr += np.count_nonzero(constraints.nonlinear_eq)
        super().__init__(
            n_var=n_var,
            n_obj=1,
            n_ieq_constr=n_ieq_constr,
            n_eq_constr=n_eq_constr,
            xl=lower,
            xu=upper,
            elementwise=parallel is False,
        )
        self._function = function
        self._constraints = constraints

    def _evaluate(
        self,
        variables: NDArray[np.float64],
        out: Dict[str, Any],
        *_0: Any,  # noqa: ANN401
        **_1: Any,  # noqa: ANN401
    ) -> None:
        variables = variables.astype(np.float64)
        out["F"] = self._function(variables)
        if self._constraints.ineq_func is not None:
            out["G"] = self._constraints.ineq_func(variables)
        if self._constraints.eq_func is not None:
            out["H"] = self._constraints.eq_func(variables)

    def __deepcopy__(self, memo: Any) -> _Problem:  # noqa: ANN401
        # Deep copy does not work on the plain object due to the callbacks.
        return _Problem(
            copy.deepcopy(self.n_var),
            lower=copy.deepcopy(self.xl),
            upper=copy.deepcopy(self.xu),
            function=self._function,
            constraints=self._constraints,
        )


class PyMooOptimizer(Optimizer):
    """Plugin class for optimization via pymoo."""

    def __init__(
        self, config: EnOptConfig, optimizer_callback: OptimizerCallback
    ) -> None:
        """Initialize the optimizer implemented by the pymoo plugin.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        self._config = config
        self._optimizer_callback = optimizer_callback
        options = (
            copy.deepcopy(self._config.optimizer.options)
            if isinstance(self._config.optimizer.options, dict)
            else {}
        )
        _, _, method = self._config.optimizer.method.rpartition("/")
        options["algorithm"] = method
        self._parameters = ParametersConfig.model_validate(options)
        self._bounds = self._get_bounds()
        self._constraints = self._get_constraints()
        self._cached_variables: Optional[NDArray[np.float64]] = None
        self._cached_function: Optional[NDArray[np.float64]] = None
        self._stdout: TextIO

        self._allow_nan = True
        for algorithm in _NO_FAILURE_HANDLING:
            if algorithm in self._parameters.algorithm:
                self._allow_nan = False

    def start(self, initial_values: NDArray[np.float64]) -> None:
        """Start the optimization.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        self._cached_variables = None
        self._cached_function = None

        variable_indices = self._config.variables.indices
        if variable_indices is not None:
            initial_values = initial_values[variable_indices]

        problem = _Problem(
            n_var=initial_values.size,
            lower=self._bounds[0],
            upper=self._bounds[1],
            constraints=self._constraints,
            function=self._calculate_objective,
            parallel=self._config.optimizer.parallel,
        )
        if self._parameters.constraints is not None:
            constraints = self._parameters.get_constraints()
            problem = constraints(problem, **self._parameters.constraints.parameters)

        output_dir = self._config.optimizer.output_dir
        output_file: Union[str, Path]
        if output_dir is None:
            output_file = os.devnull
        else:
            output_file = create_output_path(_OUTPUT_FILE, output_dir, suffix=".txt")

        self._stdout = sys.stdout

        with Path(output_file).open("a", encoding="utf-8") as output, redirect_stdout(
            output
        ):
            minimize(
                problem,
                self._parameters.get_algorithm(),
                termination=self._parameters.get_termination(),
                seed=self._parameters.seed,
                verbose=output_dir is not None,
            )

    @property
    def allow_nan(self) -> bool:
        """Whether NaN is allowed.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        return self._allow_nan

    def _get_bounds(self) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        lower_bounds = self._config.variables.lower_bounds
        upper_bounds = self._config.variables.upper_bounds
        variable_indices = self._config.variables.indices
        if variable_indices is not None:
            lower_bounds = lower_bounds[variable_indices]
            upper_bounds = upper_bounds[variable_indices]
        return lower_bounds, upper_bounds

    def _get_constraints(self) -> _Constraints:
        constraints = _Constraints()

        nonlinear_config = self._config.nonlinear_constraints
        if nonlinear_config is not None:
            constraints.nonlinear_eq = nonlinear_config.types == ConstraintType.EQ
            constraints.nonlinear_ineq = nonlinear_config.types != ConstraintType.EQ

        linear_config = self._config.linear_constraints
        if linear_config is not None:
            if self._config.variables.indices is not None:
                linear_config = filter_linear_constraints(
                    linear_config, self._config.variables.indices
                )
            constraints.linear_eq = linear_config.types == ConstraintType.EQ
            constraints.linear_ineq = linear_config.types != ConstraintType.EQ
            constraints.coefficients = linear_config.coefficients.copy()
            constraints.rhs_values = linear_config.rhs_values.copy()
            constraints.coefficients[linear_config.types == ConstraintType.GE] *= -1.0
            constraints.rhs_values[linear_config.types == ConstraintType.GE] *= -1.0

        if np.any(constraints.nonlinear_ineq) or np.any(constraints.linear_ineq):
            constraints.ineq_func = partial(
                self._calculate_constraints,
                nonlinear=constraints.nonlinear_ineq,
                linear=constraints.linear_ineq,
            )
        if np.any(constraints.nonlinear_eq) or np.any(constraints.linear_eq):
            constraints.eq_func = partial(
                self._calculate_constraints,
                nonlinear=constraints.nonlinear_eq,
                linear=constraints.linear_eq,
            )

        return constraints

    def _calculate_objective(
        self, variables: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        functions = self._get_functions(variables)
        if variables.ndim > 1:
            return functions[:, 0]
        return np.array(functions[0])

    def _calculate_constraints(
        self,
        variables: NDArray[np.float64],
        nonlinear: NDArray[np.bool_],
        linear: NDArray[np.bool_],
    ) -> NDArray[np.float64]:
        have_nonlinear = np.any(nonlinear)
        have_linear = np.any(linear)
        if have_nonlinear:
            functions = self._get_functions(variables)
            nonlinear_constraints = (
                functions[1:][nonlinear]
                if variables.ndim == 1
                else functions[:, 1:][:, nonlinear]
            )
        if have_linear:
            coeffs = self._constraints.coefficients[linear]
            rhs = self._constraints.rhs_values[linear]
            linear_constraints = (
                np.array(np.matmul(coeffs, variables) - rhs)
                if variables.ndim == 1
                else np.vstack(
                    [
                        np.matmul(coeffs, variables[idx, :]) - rhs
                        for idx in range(variables.shape[0])
                    ],
                )
            )
        if have_nonlinear and have_linear:
            return np.hstack((nonlinear_constraints, linear_constraints))
        if have_nonlinear:
            return nonlinear_constraints
        return linear_constraints

    def _get_functions(self, variables: NDArray[np.float64]) -> NDArray[np.float64]:
        if (
            self._cached_variables is None
            or variables.shape != self._cached_variables.shape
            or not np.allclose(variables, self._cached_variables)
        ):
            self._cached_variables = None
            self._cached_function = None
        if self._cached_function is None:
            self._cached_variables = variables.copy()
            with redirect_stdout(self._stdout):
                function, _ = self._optimizer_callback(
                    variables,
                    return_functions=True,
                    return_gradients=False,
                )
                if self._allow_nan:
                    function = np.where(np.isnan(function), np.inf, function)
            self._cached_function = function.copy()
        return self._cached_function


class PyMooOptimizerPlugin(OptimizerPlugin):
    """Default filter transform plugin class."""

    def create(
        self, config: EnOptConfig, optimizer_callback: OptimizerCallback
    ) -> PyMooOptimizer:
        """Initialize the optimizer plugin.

        See the [ropt.plugins.optimizer.base.OptimizerPlugin][] abstract base class.

        # noqa
        """
        return PyMooOptimizer(config, optimizer_callback)

    def is_supported(self, method: str) -> bool:
        """Check if a method is supported.

        See the [ropt.plugins.optimizer.base.OptimizerPlugin][] abstract base class.

        # noqa
        """
        module_name, _, class_name = method.rpartition(".")
        if not module_name:
            return False
        full_module_name = f"pymoo.algorithms.{module_name}"
        try:
            module = importlib.import_module(full_module_name)
        except ImportError:
            return False
        for _, cls in inspect.getmembers(module, inspect.isclass):
            if cls.__name__ == class_name:
                return True
        return False
